
ACCEPTABLE_STALE_HOURS = 48

AUCTION_START_GAS_CONSTANT = 25000
AUCTION_START_GAS_MARGINAL = 39000

EMPTY_SHA3_BYTES = b'\0' * 32
EMPTY_ADDR_HEX = '0x' + '00' * 20

MIN_ETH_LABEL_LENGTH = 7

REVERSE_REGISTRAR_DOMAIN = 'addr.reverse'
